/*	FBullCowGame class file. This class will be the game itself, with its due
	variables and methods. Direct interactions with the player is delegated
	to main.cpp.
*/

#pragma once
#include "FBullCowGame.h"
#include <map>
#include <string>

// Make syntax Unreal friendly
#define TMap std::map
using FString = std::string;
using int32 = int;



/*	Map of hidden words, given the selected length. The user can select
	the game's difficulty by choosing longer words.
*/
TMap<FString, FString> WordLengthToHiddenWord{
		{"3","dog"},
		{"4","bike"},
		{"5","quite"},
		{"6","borrow"},
		{"7","pikachu"}
};

FBullCowGame::FBullCowGame() {	Constructor(); }		// default constructor

int32 FBullCowGame::GetCurrentTry() const { return MyCurrentTry; }
int32 FBullCowGame::GetWordLength() const { return MyHiddenWord.length(); }

// sets dummy values for first initialization of the game instance
void FBullCowGame::Constructor()
{
	MyHiddenWord = "";
	MyCurrentTry = 0;
	MyGuess = "";
	MyGuessValidity = false;
	MyBCCount.MyBulls = 0;
	MyBCCount.MyCows = 0;
	return;
}


FWordLengthRange FBullCowGame::GetWordLengthRange() const
{
	// HACK access actual values in WordLengthToHiddenWord in the future, instead of hard coding them here
	FWordLengthRange MyWLRange;
	MyWLRange.MyWLMin = 3;
	MyWLRange.MyWLMax = 7;
	return MyWLRange;
}

EDifficultyStatus FBullCowGame::CheckDifficultyValidity(FString Difficulty) const
{
	std::size_t FoundForbidden = Difficulty.find_first_not_of("1234567890");
	if (FoundForbidden != std::string::npos || Difficulty.length() == 0){
		return EDifficultyStatus::Not_Number;
	}
	if (WordLengthToHiddenWord[Difficulty].length() == 0) {		// if it is not a number within the range,
		return EDifficultyStatus::Out_of_Bounds;
	}
	return EDifficultyStatus::OK;
}

// receives a VALID ("3"-"7" string) difficulty, sets the game hidden word and the current try to 1
void FBullCowGame::Set(FString Difficulty)
{
	MyCurrentTry = 1;
	MyHiddenWord = WordLengthToHiddenWord[Difficulty];
	return;
}

int32 FBullCowGame::GetMaxTries() const
{
	// maximum number of tries are returned from this hard-coded map
	std::map<int, int> WordLengthToMaxTries{ {3, 4}, {4, 7}, {5, 10}, {6, 16}, {7, 20} };
	return WordLengthToMaxTries[MyHiddenWord.length()];
}

bool FBullCowGame::IsLowercase(FString Guess) const
{
	for (auto Letter : Guess) {
		if (!islower(Letter)) {
			return false;
		}
	}
	return true;
}

EGuessStatus FBullCowGame::CheckGuessValidity(FString Guess) const
{
	if (Guess.length() != GetWordLength())
	{
		return EGuessStatus::Wrong_Size;
	} 
	else if (HasForbidChar(Guess))
	{
		return EGuessStatus::Forbidden_Characters;
	}
	else if (!IsIsogram(Guess))
	{
		return EGuessStatus::Not_Isogram;
	}
	else 
	{
		return EGuessStatus::OK;
	}
}

// checks if Guess is an isogram by hash table of already-seen letters (efficient)
bool FBullCowGame::IsIsogram(FString Guess) const
{	
	// strings of length 0 and 1 should be considered as isograms
	// strings with non-repeating characters are considered isograms (even forbid. chars.)
	TMap<char, bool> LetterSeen;
	for (auto Letter : Guess) {
		if (LetterSeen[Letter]) {
			return false;
		}
		else {
			LetterSeen[Letter] = true;
		}
	}
	return true;
}

bool FBullCowGame::HasForbidChar(FString Guess) const
{
	std::size_t found = Guess.find_first_not_of("abcdefghijklmnopqrstuvwxyz");
	if (found != std::string::npos)
	{
		return true;
	}
	else {
		return false;
	}
}

void FBullCowGame::IncreaseCurrentTry()
{
	MyCurrentTry++;
	return;
}

// receives a VALID guess and counts how many Bulls and Cows were scored
FBullCowCount FBullCowGame::CountBullsCows(FString Guess)
{
	MyBCCount.MyBulls = 0;
	MyBCCount.MyCows = 0;
	for(int32 GChar = 0; GChar != GetWordLength(); GChar++) {
		for (int32 MHWChar = 0; MHWChar != GetWordLength(); MHWChar++) {
			if (Guess[GChar] == MyHiddenWord[MHWChar]) {
				if (GChar == MHWChar) {
					MyBCCount.MyBulls++;
				}
				else {
					MyBCCount.MyCows++;
				}
			}
		}
	}
	return MyBCCount;
}

// game is won when all input characters are bulls
bool FBullCowGame::IsGameWon() const
{
	if (MyBCCount.MyBulls == GetWordLength()) {
		return true;
	} else {
		return false;
	}
	
}

/* old IsIsogram method (direct comparison)
bool FBullCowGame::IsIsogram(FString Guess) const
{
	// loops though guess characters
	for (int32 i = 0; i != GetWordLength() - 1; i++) {
		// loops through guess characters again
		for (int32 j = i + 1; j != GetWordLength(); j++) {
			// if the two characters are equal, return false
			if (Guess[i] == Guess[j]) {
				return false;
			}
		}
	}
	// if no coincidences were found, return true
	return true;

}
*/
