/*	Header file for the Bulls and Cows game class, with its due variables and method.
	The game is a simple guess-the-word game, based on Mastermind.
	The game class handles internal variables and processing of user inputs. Direct interaction
	with the player is handled in main.cpp.
*/

#pragma once
#include <string>

// Make syntax Unreal friendly
using FString = std::string;
using int32 = int;



/*	Class to handle difficulty request (number of letters in the hidden word) checking.
	Depending on user input, the request could be invalid for a few different reasons.
	Each reason will give a different status, which is then used to give a sensible to
	the player and steer him/her to inputting a valid request.
*/
enum class EDifficultyStatus {
	OK,
	Not_Number,
	Out_of_Bounds,
};

/*	Struct to handle minimum and maximum word-lengths that are supported for the game.*/
struct FWordLengthRange
{
	int32 MyWLMax = 0;
	int32 MyWLMin = 0;
};

/*	Class to handle the checking of a player guess.
	Depending on user input, the request could be invalid for a few different reasons.
	Each reason will give a different status, which is then used to give a sensible to
	the player and steer him/her to inputting a valid request.
*/
enum class EGuessStatus
{
	OK,
	Not_Isogram,
	Forbidden_Characters,
	Wrong_Size
};

/*	Struct to handle the number of bulls and cows scored by a given (valid) guess.*/
struct FBullCowCount
{
	int32 MyBulls = 0;
	int32 MyCows = 0;
};

/*	This class will be the game itself, with its due variables and methods.
	It stores as variables: the current try, the number of bulls and cows, the current
	guess, the hidden word and the guess validity status.
	It has methods to process user inputs (check guess validity, count bulls and cows,
	increase try number), to set the hidden word (as requested by user), and so forth.

	The interaction with the player is delegated to the main.cpp code. Everything else
	is in here.
*/
class FBullCowGame
{
public:
	FBullCowGame(); // constructor

	int32 GetCurrentTry() const;
	int32 GetWordLength() const;
	int32 GetMaxTries() const;

	FWordLengthRange GetWordLengthRange() const;
	EDifficultyStatus CheckDifficultyValidity(FString) const;
	EGuessStatus CheckGuessValidity(FString) const;
	bool IsIsogram(FString) const;
	bool IsLowercase(FString)const;
	bool HasForbidChar(FString) const;
	bool IsGameWon() const;	

	void Constructor();
	void Set(FString);
	void IncreaseCurrentTry();
	FBullCowCount CountBullsCows(FString);
	

private:
	// initializations in constructor
	FString MyHiddenWord;
	int32 MyCurrentTry;
	FString MyGuess;
	bool MyGuessValidity;
	FBullCowCount MyBCCount;
};